@@ -149,7 +149,7 @@ loop(int toplevel, int justonce)
 	    continue;
 	}
 	if (hend(prog)) {
-	    int toksav = tok;
+	    enum lextok toksav = tok;
 
 	    non_empty = 1;
 	    if (toplevel &&
@@ -215,6 +215,7 @@ loop(int toplevel, int justonce)
     return LOOP_OK;
 }
 
+/* Shared among parseargs(), parseopts(), init_io(), and init_misc() */
 static char *cmd;
 static int restricted;
 
@@ -222,9 +223,7 @@ static int restricted;
 static void
 parseargs(char **argv, char **runscript)
 {
-    int optionbreak = 0;
     char **x;
-    int action, optno;
     LinkList paramlist;
 
     argzero = *argv++;
@@ -247,93 +246,10 @@ parseargs(char **argv, char **runscript)
     opts[SHINSTDIN] = 0;
     opts[SINGLECOMMAND] = 0;
 
-    /* loop through command line options (begins with "-" or "+") */
-    while (!optionbreak && *argv && (**argv == '-' || **argv == '+')) {
-	char *args = *argv;
-	action = (**argv == '-');
-	if (!argv[0][1])
-	    *argv = "--";
-	while (*++*argv) {
-	    if (**argv == '-') {
-		if(!argv[0][1]) {
-		    /* The pseudo-option `--' signifies the end of options. */
-		    argv++;
-		    goto doneoptions;
-		}
-		if(*argv != args+1 || **argv != '-')
-		    goto badoptionstring;
-		/* GNU-style long options */
-		++*argv;
-		if (!strcmp(*argv, "version")) {
-		    printf("zsh %s (%s-%s-%s)\n",
-			    ZSH_VERSION, MACHTYPE, VENDOR, OSTYPE);
-		    exit(0);
-		}
-		if (!strcmp(*argv, "help")) {
-		    printhelp();
-		    exit(0);
-		}
-		/* `-' characters are allowed in long options */
-		for(args = *argv; *args; args++)
-		    if(*args == '-')
-			*args = '_';
-		goto longoptions;
-	    }
+    if (parseopts(NULL, &argv, opts, &cmd, NULL))
+	exit(1);
 
-	    if (unset(SHOPTIONLETTERS) && **argv == 'b') {
-		/* -b ends options at the end of this argument */
-		optionbreak = 1;
-	    } else if (**argv == 'c') {
-		/* -c command */
-		cmd = *argv;
-		opts[INTERACTIVE] &= 1;
-		scriptname = scriptfilename = ztrdup("zsh");
-	    } else if (**argv == 'o') {
-		if (!*++*argv)
-		    argv++;
-		if (!*argv) {
-		    zerr("string expected after -o");
-		    exit(1);
-		}
-	    longoptions:
-		if (!(optno = optlookup(*argv))) {
-		    zerr("no such option: %s", *argv);
-		    exit(1);
-		} else if (optno == RESTRICTED)
-		    restricted = action;
-		else
-		    dosetopt(optno, action, 1);
-              break;
-	    } else if (isspace(STOUC(**argv))) {
-		/* zsh's typtab not yet set, have to use ctype */
-		while (*++*argv)
-		    if (!isspace(STOUC(**argv))) {
-		    badoptionstring:
-			zerr("bad option string: `%s'", args);
-			exit(1);
-		    }
-		break;
-	    } else {
-	    	if (!(optno = optlookupc(**argv))) {
-		    zerr("bad option: -%c", **argv);
-		    exit(1);
-		} else if (optno == RESTRICTED)
-		    restricted = action;
-		else
-		    dosetopt(optno, action, 1);
-	    }
-	}
-	argv++;
-    }
-    doneoptions:
     paramlist = znewlinklist();
-    if (cmd) {
-	if (!*argv) {
-	    zerr("string expected after -%s", cmd);
-	    exit(1);
-	}
-	cmd = *argv++;
-    }
     if (*argv) {
 	if (unset(SHINSTDIN)) {
 	    if (cmd)
@@ -361,6 +277,165 @@ parseargs(char **argv, char **runscript)
     argzero = ztrdup(argzero);
 }
 
+/* Insert into list in order of pointer value */
+
+/**/
+static void
+parseopts_insert(LinkList optlist, char *base, int optno)
+{
+    LinkNode node;
+    void *ptr = base + (optno < 0 ? -optno : optno);
+
+    for (node = firstnode(optlist); node; incnode(node)) {
+	if (ptr < getdata(node)) {
+	    insertlinknode(optlist, prevnode(node), ptr);
+	    return;
+	}
+    }
+
+    addlinknode(optlist, ptr);
+}
+
+/*
+ * Parse shell options.
+ * If nam is not NULL, this is called from a command; don't
+ * exit on failure.
+ *
+ * If optlist is not NULL, it used to form a list of pointers
+ * into new_opts indicating which options have been changed.
+ */
+
+/**/
+mod_export int
+parseopts(char *nam, char ***argvp, char *new_opts, char **cmdp,
+	  LinkList optlist)
+{
+    int optionbreak = 0;
+    int action, optno;
+    char **argv = *argvp;
+
+    *cmdp = 0;
+#define WARN_OPTION(F, S)						\
+    do {								\
+	if (nam)							\
+	    zwarnnam(nam, F, S);					\
+	else								\
+	    zerr(F, S);							\
+    } while (0)
+#define LAST_OPTION(N)	       \
+    do {		       \
+	if (nam) {	       \
+	    if (*argv)	       \
+		argv++;	       \
+	    goto doneargv;     \
+	} else exit(N);	       \
+    } while(0)
+
+    /* loop through command line options (begins with "-" or "+") */
+    while (!optionbreak && *argv && (**argv == '-' || **argv == '+')) {
+	char *args = *argv;
+	action = (**argv == '-');
+	if (!argv[0][1])
+	    *argv = "--";
+	while (*++*argv) {
+	    if (**argv == '-') {
+		if(!argv[0][1]) {
+		    /* The pseudo-option `--' signifies the end of options. */
+		    argv++;
+		    goto doneoptions;
+		}
+		if(*argv != args+1 || **argv != '-')
+		    goto badoptionstring;
+		/* GNU-style long options */
+		++*argv;
+		if (!strcmp(*argv, "version")) {
+		    printf("zsh %s (%s-%s-%s)\n",
+			    ZSH_VERSION, MACHTYPE, VENDOR, OSTYPE);
+		    LAST_OPTION(0);
+		}
+		if (!strcmp(*argv, "help")) {
+		    printhelp();
+		    LAST_OPTION(0);
+		}
+		/* `-' characters are allowed in long options */
+		for(args = *argv; *args; args++)
+		    if(*args == '-')
+			*args = '_';
+		goto longoptions;
+	    }
+
+	    if (unset(SHOPTIONLETTERS) && **argv == 'b') {
+		/* -b ends options at the end of this argument */
+		optionbreak = 1;
+	    } else if (**argv == 'c') {
+		/* -c command */
+		*cmdp = *argv;
+		new_opts[INTERACTIVE] &= 1;
+		scriptname = scriptfilename = ztrdup("zsh");
+	    } else if (**argv == 'o') {
+		if (!*++*argv)
+		    argv++;
+		if (!*argv) {
+		    WARN_OPTION("string expected after -o", NULL);
+		    return 1;
+		}
+	    longoptions:
+		if (!(optno = optlookup(*argv))) {
+		    WARN_OPTION("no such option: %s", *argv);
+		    return 1;
+		} else if (optno == RESTRICTED && !nam) {
+		    restricted = action;
+		} else if ((optno == EMACSMODE || optno == VIMODE) && nam) {
+		    WARN_OPTION("can't change option: %s", *argv);
+		} else {
+		    if (dosetopt(optno, action, !nam, new_opts) && nam) {
+			WARN_OPTION("can't change option: %s", *argv);
+		    } else if (optlist) {
+			parseopts_insert(optlist, new_opts, optno);
+		    }
+		}
+              break;
+	    } else if (isspace(STOUC(**argv))) {
+		/* zsh's typtab not yet set, have to use ctype */
+		while (*++*argv)
+		    if (!isspace(STOUC(**argv))) {
+		     badoptionstring:
+			WARN_OPTION("bad option string: '%s'", args);
+			return 1;
+		    }
+		break;
+	    } else {
+	    	if (!(optno = optlookupc(**argv))) {
+		    WARN_OPTION("bad option: -%c", **argv);
+		    return 1;
+		} else if (optno == RESTRICTED && !nam) {
+		    restricted = action;
+		} else if ((optno == EMACSMODE || optno == VIMODE) && nam) {
+		    WARN_OPTION("can't change option: %s", *argv);
+		} else {
+		    if (dosetopt(optno, action, !nam, new_opts) && nam) {
+			WARN_OPTION("can't change option: -%c", **argv);
+		    } else if (optlist) {
+			parseopts_insert(optlist, new_opts, optno);
+		    }
+		}
+	    }
+	}
+	argv++;
+    }
+ doneoptions:
+    if (*cmdp) {
+	if (!*argv) {
+	    WARN_OPTION("string expected after -%s", *cmdp);
+	    return 1;
+	}
+	*cmdp = *argv++;
+    }
+ doneargv:
+    *argvp = argv;
+    return 0;
+}
+
 /**/
 static void
 printhelp(void)
@@ -1039,6 +1114,7 @@ init_signals(void)
     install_handler(SIGCHLD);
 #ifdef SIGWINCH
     install_handler(SIGWINCH);
+    winch_block();	/* See utils.c:preprompt() */
 #endif
     if (interact) {
 	install_handler(SIGALRM);
@@ -1137,7 +1213,7 @@ init_misc(void)
 #else
     if (*zsh_name == 'r' || restricted)
 #endif
-	dosetopt(RESTRICTED, 1, 0);
+	dosetopt(RESTRICTED, 1, 0, opts);
     if (cmd) {
 	if (SHIN >= 10)
 	    fclose(bshin);
@@ -1200,7 +1276,7 @@ source(char *s)
     subsh  = 0;
     lineno = 1;
     loops  = 0;
-    dosetopt(SHINSTDIN, 0, 1);
+    dosetopt(SHINSTDIN, 0, 1, opts);
     scriptname = s;
     scriptfilename = s;
 
@@ -1272,7 +1348,7 @@ source(char *s)
     thisjob = cj;                    /* current job number                   */
     lineno = oldlineno;              /* our current lineno                   */
     loops = oloops;                  /* the # of nested loops we are in      */
-    dosetopt(SHINSTDIN, oldshst, 1); /* SHINSTDIN option                     */
+    dosetopt(SHINSTDIN, oldshst, 1, opts); /* SHINSTDIN option               */
     errflag = 0;
     if (!exit_pending)
 	retflag = 0;
@@ -1508,9 +1584,10 @@ zsh_main(UNUSED(int argc), char **argv)
 
     fdtable_size = zopenmax();
     fdtable = zshcalloc(fdtable_size*sizeof(*fdtable));
+    fdtable[0] = fdtable[1] = fdtable[2] = FDT_EXTERNAL;
 
     createoptiontable();
-    emulate(zsh_name, 1);   /* initialises most options */
+    emulate(zsh_name, 1, &emulation, opts);   /* initialises most options */
     opts[LOGINSHELL] = (**argv == '-');
     opts[PRIVILEGED] = (getuid() != geteuid() || getgid() != getegid());
     opts[USEZLE] = 1;   /* may be unset in init_io() */
@@ -1534,15 +1611,20 @@ zsh_main(UNUSED(int argc), char **argv)
 	 * We only do this at top level, because if we are
 	 * executing stuff we may refer to them by job pointer.
 	 */
+	int errexit = 0;
 	maybeshrinkjobtab();
 
 	do {
 	    /* Reset return from top level which gets us back here */
 	    retflag = 0;
 	    loop(1,0);
+	    if (errflag && !interact && !isset(CONTINUEONERROR)) {
+		errexit = 1;
+		break;
+	    }
 	} while (tok != ENDINPUT && (tok != LEXERR || isset(SHINSTDIN)));
-	if (tok == LEXERR) {
-	    /* Make sure a parse error exits with non-zero status */
+	if (tok == LEXERR || errexit) {
+	    /* Make sure a fatal error exits with non-zero status */
 	    if (!lastval)
 		lastval = 1;
 	    stopmsg = 1;
